# *_*coding:utf-8 *_*
from os.path import dirname, abspath, join

ROOT = dirname(abspath(__file__))
import click
from click_help_colors import HelpColorsGroup

# CONSTANTS
CLICK_CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'])
CLICK_CONTEXT_SETTINGS_NO_HELP = dict(help_option_names=[])


def print_version(ctx, param, value):
    if not value or ctx.resilient_parsing:
        return
    package_conf = {}
    with open(join(ROOT, "__version__.py")) as f:
        exec(f.read(), package_conf)
    # click.secho(package_conf['__version__'], fg='green')
    click.secho(package_conf['__version__'], blink=True, bold=True)
    ctx.exit()


@click.group(cls=HelpColorsGroup,
             context_settings=CLICK_CONTEXT_SETTINGS,
             help_headers_color='yellow',
             help_options_color='magenta',
             help_options_custom_colors={
                 'version': 'green',
             })
@click.option('-v',
              '--version',
              is_flag=True,
              callback=print_version,
              expose_value=False,
              is_eager=True,
              help='Show version')
def cli():
    """\b
██╗      █████╗ ███╗   ██╗ ██████╗ ███╗   ███╗ █████╗
██║     ██╔══██╗████╗  ██║██╔════╝ ████╗ ████║██╔══██╗
██║     ███████║██╔██╗ ██║██║  ███╗██╔████╔██║███████║
██║     ██╔══██║██║╚██╗██║██║   ██║██║╚██╔╝██║██╔══██║
███████╗██║  ██║██║ ╚████║╚██████╔╝██║ ╚═╝ ██║██║  ██║
╚══════╝╚═╝  ╚═╝╚═╝  ╚═══╝ ╚═════╝ ╚═╝     ╚═╝╚═╝  ╚═╝
    """


def run():
    try:
        cli()
    except Exception as e:
        pass


if __name__ == "__main__":
    run()



"""
Visualize insights and statistics of a text-based Pandas DataFrame.
"""

import pandas as pd
import numpy as np
import plotly.express as px

from wordcloud import WordCloud

from langma.utils._types import TextSeries, InputSeries
import string
import warnings

from matplotlib.colors import LinearSegmentedColormap as lsg
import matplotlib.pyplot as plt
from typing import Dict, List
from collections import Counter

plt.rcParams['font.sans-serif'] = ['SimHei']
plt.rcParams['axes.unicode_minus'] = False
plt.style.use('ggplot')

warnings.filterwarnings('ignore')
warnings.simplefilter('ignore')


def scatterplot(
    df: pd.DataFrame,
    col: str,
    color: str = None,
    hover_name: str = None,
    hover_data: [] = None,
    title="",
    return_figure=False,
):
    """
    Show scatterplot of DataFrame column using python plotly scatter.

    Plot the values in column col. For example, if every cell in df[col]
    is a list of three values (e.g. from doing PCA with 3 components),
    a 3D-Plot is created and every cell entry [x, y, z] is visualized
    as the point (x, y, z).

    Parameters
    ----------
    df: DataFrame with a column to be visualized.

    col: str
        The name of the column of the DataFrame to use for x and y (and z)
        axis.

    color: str, optional, default=None
        Name of the column to use for coloring (rows with same value get same
        color).

    hover_name: str, optional, default=None
        Name of the column to supply title of hover data when hovering over a
        point.

    hover_data: List[str], optional, default=[]
        List of column names to supply data when hovering over a point.

    title: str, default to "".
        Title of the plot.

    return_figure: bool, optional, default=False
        Function returns the figure instead of showing it if set to True.

    Examples
    --------
    >>> import texthero as hero
    >>> import pandas as pd
    >>> df = pd.DataFrame(["Football, Sports, Soccer",
    ...                    "music, violin, orchestra", "football, fun, sports",
    ...                    "music, fun, guitar"], columns=["texts"])
    >>> df["texts"] = hero.clean(df["texts"]).pipe(hero.tokenize)
    >>> df["pca"] = (
    ...             hero.tfidf(df["texts"])
    ...                 .pipe(hero.pca, n_components=3)
    ... )
    >>> df["topics"] = (
    ...                hero.tfidf(df["texts"])
    ...                    .pipe(hero.kmeans, n_clusters=2)
    ... )
    >>> hero.scatterplot(df, col="pca", color="topics",
    ...                  hover_data=["texts"]) # doctest: +SKIP
    """

    plot_values = np.stack(df[col], axis=1)
    dimension = len(plot_values)

    if dimension < 2 or dimension > 3:
        raise ValueError(
            "The column you want to visualize has dimension < 2 or dimension > 3."
            " The function can only visualize 2- and 3-dimensional data.")

    if dimension == 2:
        x, y = plot_values[0], plot_values[1]

        fig = px.scatter(
            df,
            x=x,
            y=y,
            color=color,
            hover_data=hover_data,
            title=title,
            hover_name=hover_name,
        )

    else:
        x, y, z = plot_values[0], plot_values[1], plot_values[2]

        fig = px.scatter_3d(
            df,
            x=x,
            y=y,
            z=z,
            color=color,
            hover_data=hover_data,
            title=title,
            hover_name=hover_name,
        )

    if return_figure:
        return fig
    else:
        fig.show()


"""
Wordcloud
"""


@InputSeries(TextSeries)
def wordcloud(
    s: TextSeries,
    font_path: str = None,
    width: int = 400,
    height: int = 200,
    max_words=200,
    mask=None,
    contour_width=0,
    contour_color="PAPAYAWHIP",
    min_font_size=4,
    background_color="PAPAYAWHIP",
    max_font_size=None,
    relative_scaling="auto",
    colormap=None,
    return_figure=False,
):
    """
    Plot wordcloud image using WordCloud from word_cloud package.

    Most of the arguments are very similar if not equal to the mother
    function. In constrast, all words are taken into account when computing
    the wordcloud, inclusive stopwords. They can be easily removed with
    preprocessing.remove_stopwords.

    Words are computed using generate_from_frequencies.

    To reduce blur in the wordcloud image, `width` and `height` should be at
    least 400.

    Parameters
    ----------
    s : :class:`texthero._types.TextSeries`

    font_path : str, optional, default=None
        Font path to the font that will be used (OTF or TTF). Defaults to
        DroidSansMono path on a Linux machine. If you are on another OS or
        don't have this font, you need to adjust this path.

    width : int, optional, default=400
        Width of the canvas.

    height : int, optional, default=200
        Height of the canvas.

    max_words : int, optional, default=200
        The maximum number of words.

    mask : nd-array or None, optional, default=None
        When set, gives a binary mask on where to draw words. When set, width
        and height will be ignored and the shape of mask will be used instead.
        All white (#FF or #FFFFFF) entries will be considerd "masked out"
        while other entries will be free to draw on.

    contour_width: float, optional, default=0
        If mask is not None and contour_width > 0, draw the mask contour.

    contour_color: str, optional, default="PAPAYAWHIP"
        Mask contour color.

    min_font_size : int, optional, default=4
        Smallest font size to use. Will stop when there is no more room in
        this size.

    background_color : str, optional, default="PAPAYAWHIP"
        Background color for the word cloud image.

    max_font_size : int or None, optional, default=None
        Maximum font size for the largest word. If None, height of the image
        is used.

    relative_scaling : float, optional, default="auto"
        Importance of relative word frequencies for font-size.  With
        relative_scaling=0, only word-ranks are considered.  With
        relative_scaling=1, a word that is twice as frequent will have twice
        the size.  If you want to consider the word frequencies and not only
        their rank, relative_scaling around .5 often looks good.
        If 'auto' it will be set to 0.5 unless repeat is true, in which
        case it will be set to 0.

    colormap : string or matplotlib colormap, optional, default="viridis"
        Matplotlib colormap to randomly draw colors from for each word.

    """
    text = s.str.cat(sep=" ")

    if colormap is None:

        # Custom palette.
        # TODO move it under tools.
        corn = (255.0 / 256, 242.0 / 256, 117.0 / 256)
        mango_tango = (255.0 / 256, 140.0 / 256, 66.0 / 256)
        crayola = (63.0 / 256, 136.0 / 256, 197.0 / 256)
        crimson = (215.0 / 256, 38.0 / 256, 61.0 / 256)
        oxford_blue = (2.0 / 256, 24.0 / 256, 43.0 / 256)

        texthero_cm = lsg.from_list(
            "texthero", [corn, mango_tango, crayola, crimson, oxford_blue])

        colormap = texthero_cm

    words = s.str.cat(sep=" ").split()

    wordcloud = WordCloud(
        font_path=font_path,
        width=width,
        height=height,
        max_words=max_words,
        mask=mask,
        contour_width=contour_width,
        contour_color=contour_color,
        min_font_size=min_font_size,
        background_color=background_color,
        max_font_size=max_font_size,
        relative_scaling=relative_scaling,
        colormap=colormap,
        # stopwords=[],  # TODO. Will use generate from frequencies.
        # normalize_plurals=False,  # TODO.
    ).generate_from_frequencies(dict(Counter(words)))

    # fig = px.imshow(wordcloud)
    # fig.show()

    fig, ax = plt.subplots(figsize=(20, 10))
    ax.imshow(wordcloud, interpolation="bilinear")
    ax.axis("off")

    if return_figure:
        return fig


@InputSeries(TextSeries)
def top_words(s: TextSeries, normalize=False) -> pd.Series:
    r"""
    Return a pandas series with index the top words and as value the count.

    Tokenization: split by space and remove all punctuations that are not
    between characters.

    Parameters
    ----------
    normalize : bool, optional, default=False.
        When set to true, return normalized values.

    Examples
    --------
    >>> import pandas as pd
    >>> import texthero as hero
    >>> s = pd.Series("one two two three three three")
    >>> hero.top_words(s)
    three    3
    two      2
    one      1
    dtype: int64

    """

    # Replace all punctuation that are NOT in-between chacarters
    # This means, they have either a non word-bounding \B, are at the start ^, or at the end $
    # As re.sub replace all and not just the matching group, add matching parenthesis to the character
    # to keep during replacement.

    # TODO replace it with tokenizer.

    pattern = (
        rf"((\w)[{string.punctuation}](?:\B|$)|(?:^|\B)[{string.punctuation}](\w))"
    )

    return (s.str.replace(
        pattern, r"\2 \3"
    )  # \2 and \3 permits to keep the character around the punctuation.
            .str.split()  # now split by space
            .explode()  # one word for each line
            .value_counts(normalize=normalize))


def visualize_train_history(history: Dict[str, List[float]],
                            fname=None,
                            figsize=(14, 7),
                            facecolor='white'):
    """Example:
    visualize_train_history(history.history)

    :param history:
    :return:
    """
    plt.figure(figsize=figsize, dpi=200, facecolor=facecolor)

    plt.subplot(1, 2, 1)  # 创建子图（行，列，活跃区）
    # 分别绘制 accuracy 和 val_accuracy 曲线
    plt.plot(history['accuracy'], label='accuracy', linewidth=1.0)
    try:
        plt.plot(history['val_accuracy'], label='val_accuracy', linewidth=1.0)
    except:
        pass
    plt.title('Accuracy')  # 添加图形标题
    plt.legend()  # 展示图例

    plt.subplot(1, 2, 2)
    plt.plot(history['loss'], label='loss', linewidth=1.0)
    try:
        plt.plot(history['val_loss'], label='val_loss', linewidth=1.0)
    except:
        pass
    plt.title('Loss')
    plt.legend()

    if fname:
        plt.imsave(fname)
    else:
        plt.show()


def visualize_multi_train_history(history_dict: Dict[str, Dict[str,
                                                               List[float]]],
                                  fname=None,
                                  figsize=(14, 8),
                                  facecolor='white'):
    """Example:

    visualize_multi_train_history(history_dict={
        'BiLSTM': history.history,
        'CNN': history_cnn.history
        })
    :param history_dict:
    :return:
    """
    plt.figure(figsize=figsize, dpi=200, facecolor=facecolor)

    plt.subplot(2, 2, 1)
    for model_name, data in history_dict.items():
        plt.plot(data['accuracy'], label=model_name, linewidth=1.0)
    plt.title('Accuracy')
    plt.legend()

    plt.subplot(2, 2, 2)
    try:
        for model_name, data in history_dict.items():
            plt.plot(data['val_accuracy'], label=model_name, linewidth=1.0)
    except:
        pass
    plt.title('Val Accuracy')
    plt.legend()

    plt.subplot(2, 2, 3)
    for model_name, data in history_dict.items():
        plt.plot(data['loss'], label=model_name, linewidth=1.0)
    plt.title('Loss')
    plt.legend()

    plt.subplot(2, 2, 4)
    try:
        for model_name, data in history_dict.items():
            plt.plot(data['val_loss'], label=model_name, linewidth=1.0)
    except:
        pass
    plt.title('Val Loss')
    plt.legend()
    if fname:
        plt.imsave(fname)
    else:
        plt.show()



# *_*coding:utf-8 *_*
from __future__ import absolute_import, division, print_function
import re


def filter_word(raw_word):
    if not re.search(r'^[\u4e00-\u9fa5_a-zA-Z0-9]+$', raw_word):
        return False
    else:
        return True


def filter_char(char):
    if not re.search(r'[\u4e00-\u9fa5_a-zA-Z0-9]', char):
        return False
    else:
        return True



# *_*coding:utf-8 *_*
from __future__ import absolute_import, division, print_function

import math
from datetime import timedelta, datetime
import pandas as pd
import numpy as np
import jieba
import re
from tensorflow.keras.preprocessing.text import Tokenizer
from tensorflow.keras.preprocessing import sequence
from sklearn.preprocessing import LabelEncoder, MultiLabelBinarizer, \
    OneHotEncoder
import pickle
from tqdm import tqdm
import warnings
from gensim.models.word2vec import LineSentence, Word2Vec
from langma.utils.logutil import logger

from langma.utils.file_utils import save_dict

warnings.filterwarnings('ignore')
warnings.simplefilter('ignore')
pd.set_option("display.max_columns", None)
# pd.set_option("display.max_rows", None)
pd.set_option("display.float_format", lambda x: "{:.2f}".format(x))


def get_max_len(data: pd.Series, up=False):
    """获得合适的最大长度值
    >>> MAX_LEN = get_max_len(df['content'])

    :param data: 待统计的数据
    :return: 最大长度值
    """

    max_lens = data.apply(lambda x: x.count(' ') + 1)
    if up:
        # 向上取整
        return math.ceil(np.mean(max_lens) + 2 * np.std(max_lens))
    else:
        # 向下取整
        return int(np.mean(max_lens) + 2 * np.std(max_lens))


def load_stopwords(file):
    """加载停用词"""
    return {line.strip() for line in open(file, encoding='utf-8').readlines()}


def remove_punctuation(sentence, only_chinese=False):
    '''删除特殊符号

Usage:
    >>> df['clean_title'] = df['title'].apply(remove_punctuation)

    :param sentence:
    :param only_chinese:
    :return:
    '''
    sentence = str(sentence)
    if sentence.strip() == '':
        return ''
    # 删除特殊符号
    sentence = re.sub(
        r'[\s+\-\!\/\[\]\{\}_,.$%^*(+\"\')]+|[:：+——()?【】“”！，。？、~@#￥%……&*（）]+|评|你好|您好',
        ' ', sentence)
    if only_chinese:
        # 删除汉字以外的所有符号
        rule = re.compile(u"[^\u4E00-\u9FA5]")
    else:
        # 删除除字母,数字，汉字以外的所有符号
        rule = re.compile(u"[^a-zA-Z0-9\u4E00-\u9FA5]")
    line = rule.sub('', sentence)
    return line


def sentence_preprocess(sentence, stopwords, only_chinese=False):
    # 删除特殊字符
    sentence = remove_punctuation(sentence, only_chinese=only_chinese)

    # 切词，默认精确模式，全模式cut参数cut_all=True
    words = jieba.cut(sentence, cut_all=False)

    # 去停用词
    words = [w for w in words if w not in stopwords]
    return words


def tokenizer_words(wds_series: pd.Series, num_words=50000):
    tokenizer = Tokenizer(num_words=num_words,
                          filters='!"#$%&()*+,-./:;<=>?@[\]^_`{|}~',
                          lower=True,
                          oov_token="<UNK>")
    token_wds = wds_series.values
    tokenizer.fit_on_texts(token_wds)
    word_index = tokenizer.word_index
    logger.info('共有 %s 个不相同的词语.' % len(word_index))
    sequences = tokenizer.texts_to_sequences(token_wds)
    return tokenizer, sequences


def padding_sequencies(tokend_sequences,
                       maxlen=180,
                       padding='post',
                       truncating='post'):
    # 填充sequence,让sequence的各个列的长度统一
    paded_sequences = sequence.pad_sequences(tokend_sequences,
                                             maxlen=maxlen,
                                             padding=padding,
                                             truncating=truncating)
    return paded_sequences


def create_data(text: pd.Series,
                label: pd.Series,
                num_words=50000,
                maxlen=128,
                file_x="",
                file_y="",
                tokenizer_pickle=""):
    '''

    :param text: df.content
    :param label: df.label
    :param num_words:
    :param maxlen:
    :return:
    '''
    # 对于label处理
    # le, y = multi_label_encoder(label)
    # le, y = label_encoder(label)
    le, y = onehot_encoder(label)

    # 对文本处理
    tokenizer, x = tokenizer_words(text, num_words=num_words)
    x = padding_sequencies(x, maxlen=maxlen)

    # 保存数据
    np.save(file_x, x)
    np.save(file_y, y)
    logger.info('已创建并保存x,y至：\n {} \n {}'.format(file_x, file_y))

    # 同时还要保存tokenizer和 multi_label_binarizer
    # 否则训练结束后无法还原把数字还原成文本
    tb = {'tokenizer': tokenizer, 'binarizer': le}
    with open(tokenizer_pickle, 'wb') as f:
        pickle.dump(tb, f)
    logger.info('已创建并保存tokenizer和binarizer至：\n {}'.format(tokenizer_pickle))
    return x, y


def load_processed_data(file_x="", file_y=""):
    x = np.load(file=file_x)
    y = np.load(file=file_y)
    return x, y


def label_statistics(label: pd.Series):
    '''统计标签的数量'''
    label_value_counts = label.value_counts()
    lbs = {'label': label_value_counts.index, 'count': label_value_counts}
    return pd.DataFrame(data=lbs).reset_index(drop=True)


def label_weights(labels: pd.Series) -> dict:
    """Statistics label proportion"""
    total = labels.count()
    value_counts = labels.value_counts()
    value_counts = dict(value_counts)
    for key, value in value_counts.items():
        value_counts[key] = round(value / total, 2)
    return value_counts


def tf_keras_label_encoder(label: pd.Series, num_class=3):
    from tensorflow.keras.utils import to_categorical
    return to_categorical(label.values, num_class)


def multi_label_encoder(label: pd.Series):
    '''多标签编码

    :param label:
    :return:
    '''

    # https://scikit-learn.org/stable/modules/generated/sklearn.preprocessing.LabelEncoder.html?highlight=labelencoder#sklearn.preprocessing.LabelEncoder
    mlb = MultiLabelBinarizer()
    result = mlb.fit_transform(label)
    logger.info(list(mlb.classes_))
    return mlb, result


def onehot_encoder(label: pd.Series):
    '''onehot 编码

    :param label:
    :return:
    '''

    # https://scikit-learn.org/stable/modules/generated/sklearn.preprocessing.OneHotEncoder.html?highlight=onehot#sklearn.preprocessing.OneHotEncoder
    oht = OneHotEncoder()
    result = oht.fit_transform(label.values.reshape(-1, 1))
    logger.info(list(oht.categories_))
    return oht, result


def label_encoder(label: pd.Series):
    '''
>>> from sklearn.preprocessing import LabelEncoder
>>> from collections import Counter
>>> import pandas as pd

>>> test_list = ['05db9164', '68fd1e64', '05db9164', '8cf07265', '05db9164',
             '68fd1e64', '5bfa8ab5', '5a9ed9b0', '05db9164', '9a89b36c',
             '68fd1e64', '8cf07265', '05db9164', '68fd1e64', '5a9ed9b0',
             '68fd1e64', '5a9ed9b0', '05db9164', '05db9164', '2d4ea12b']

# 初始化
>>> lbe = LabelEncoder()
# 先fit转换为set, 然后在transform。
>>> lbe_res = lbe.fit_transform(test_list)
# 如何看标签个数及顺序？[注：这里标签顺序 不是根据词频]
>>> print(lbe.classes_)
    ['05db9164' '2d4ea12b' '5a9ed9b0' '5bfa8ab5' '68fd1e64' '8cf07265' '9a89b36c']
# 如何看编码结果？
>>> print(lbe_res)
    [0 4 0 5 0 4 3 2 0 6 4 5 0 4 2 4 2 0 0 1]
# 如何看某个编码对应的原始值？
>>> print(lbe.inverse_transform([5, 2, 1]))
    ['8cf07265', '5a9ed9b0', '2d4ea12b']
    :param label:
    :return:
    '''
    # https://scikit-learn.org/stable/modules/generated/sklearn.preprocessing.LabelEncoder.html?highlight=labelencoder#sklearn.preprocessing.LabelEncoder
    le = LabelEncoder()
    result = le.fit_transform(label)
    logger.info(list(le.classes_))
    return le, result


def save_tokenizer(tokenizer, tokenizer_pickle=""):
    '''keep tokenizer'''
    with open(tokenizer_pickle, 'wb') as f:
        pickle.dump(tokenizer, f)
    logger.info('已创建并保存tokenizer和binarizer至：\n {}'.format(tokenizer_pickle))


def load_tokenizer(tokenizer_pickle=""):
    '''
    tokenizer = load_tokenizer()
    tokenizer.word_index
    :return:
    '''
    with open(tokenizer_pickle, 'rb') as f:
        tokenizer = pickle.load(f)
    return tokenizer


def load_tokenizer_binarizer(tokenizer_pickle=""):
    '''
    tokenizer, mlb = load_tokenizer_binarizer()
    tokenizer.word_index
    :return:
    '''
    with open(tokenizer_pickle, 'rb') as f:
        tb = pickle.load(f)
    return tb['tokenizer'], tb['binarizer']


def train_word2vec(seg_path,
                   train_x_pad_path,
                   train_y_pad_path,
                   test_x_pad_path,
                   embedding_file,
                   vocab_file,
                   reverse_vocab_file,
                   model_file,
                   embedding_dim=300,
                   train_epochs=3):
    logger.info('start build w2v model')
    wv_model = Word2Vec(LineSentence(seg_path),
                        size=embedding_dim,
                        sg=1,
                        workers=8,
                        iter=train_epochs,
                        window=5,
                        min_count=5)
    vocab = wv_model.wv.vocab

    # 词向量再次训练
    logger.info('start retrain w2v model')
    wv_model.build_vocab(LineSentence(train_x_pad_path), update=True)
    wv_model.train(LineSentence(train_x_pad_path),
                   epochs=1,
                   total_examples=wv_model.corpus_count)

    logger.info('1/3')
    wv_model.build_vocab(LineSentence(train_y_pad_path), update=True)
    wv_model.train(LineSentence(train_y_pad_path),
                   epochs=1,
                   total_examples=wv_model.corpus_count)

    logger.info('2/3')
    wv_model.build_vocab(LineSentence(test_x_pad_path), update=True)
    wv_model.train(LineSentence(test_x_pad_path),
                   epochs=1,
                   total_examples=wv_model.corpus_count)

    # 保存词向量模型
    wv_model.save(model_file)
    logger.info('finish retrain w2v model')
    logger.info('final w2v_model has vocabulary of ', len(wv_model.wv.vocab))

    # 12. 更新vocab
    vocab = {word: index for index, word in enumerate(wv_model.wv.index2word)}
    reverse_vocab = {
        index: word
        for index, word in enumerate(wv_model.wv.index2word)
    }

    # 保存字典
    save_dict(vocab_file, vocab)
    save_dict(reverse_vocab_file, reverse_vocab)

    # 13. 保存词向量矩阵
    embedding_matrix = wv_model.wv.vectors
    np.save(embedding_file, embedding_matrix)



# *_*coding:utf-8 *_*
from __future__ import absolute_import, division, print_function
from langma.utils.logutil import logger


def check_gpu_tf():
    import tensorflow as tf

    logger.info(tf.__version__)

    logger.info(f'GPU可用：{tf.config.list_physical_devices("GPU")}')


def check_gpu_torch():
    import torch

    logger.info(torch.__version__)

    logger.info(f'GPU可用：{torch.cuda.is_available()}')


if __name__ == '__main__':
    check_gpu_tf()
    # check_gpu_torch()



# *_*coding:utf-8 *_*
import tensorflow as tf

from langma.embeddings.abc_embedding import ABCEmbedding


class GRUDecoder(tf.keras.Model):
    def __init__(self, embedding: ABCEmbedding, hidden_size: int,
                 vocab_size: int):
        super(GRUDecoder, self).__init__()
        self.embedding = embedding

        self.gru = tf.keras.layers.GRU(hidden_size,
                                       return_sequences=True,
                                       return_state=True,
                                       recurrent_initializer='glorot_uniform')
        self.fc = tf.keras.layers.Dense(vocab_size)

    def call(self, dec_input, dec_hidden, enc_output):
        # x 在通过嵌入层后的形状 == （批大小，1，嵌入维度）
        decoder_embedding = self.embedding.embed_model(dec_input)

        s = self.gru(decoder_embedding, initial_state=dec_hidden)
        decoder_outputs, decoder_state = s

        # 输出的形状 == （批大小 * 1，隐藏层大小）
        output = tf.reshape(decoder_outputs, (-1, decoder_outputs.shape[2]))

        # 输出的形状 == （批大小，vocab）
        x = self.fc(output)
        return x, decoder_state, None
